<!doctype html>
<html>
    <head>
        <title>{% block title %}Experiment{% endblock %}</title>
        {% block head %}
        {% endblock %}
        {% block replace_stylesheets %}
            <link rel="stylesheet" href="{{ url_for('static', filename='css/bootstrap.min.css') }}" type="text/css">
            <link rel="stylesheet" href="{{ url_for('static', filename='css/dallinger.css') }}" type="text/css">
            {% block stylesheets %}
            {% endblock %}
        {% endblock %}
    </head>
    <body>
        {% block body %}
        {% endblock %}
        <div id="spinner" />
        </div>
        {% block libs %}
            <script src="{{ url_for('static', filename='scripts/jquery-3.6.0.min.js') }}" type="text/javascript"></script>
            <script src="{{ url_for('static', filename='scripts/reqwest.min.js') }}" type="text/javascript"></script>
            <script src="{{ url_for('static', filename='scripts/reconnecting-websocket.js') }}"></script>
            <script src="{{ url_for('static', filename='scripts/spin.min.js') }}" type="text/javascript"></script>
            <script src="{{ url_for('static', filename='scripts/store+json2.min.js') }}" type="text/javascript"></script>
            <script src="{{ url_for('static', filename='scripts/fingerprintjs2/1.5.1/fingerprint2.min.js') }}" type="text/javascript"></script>
            <script src="{{ url_for('static', filename='scripts/dallinger2.js') }}" type="text/javascript"> </script>
        {% endblock %}
        {% block scripts %}
            <script type="text/javascript">
                dallinger.hasAdBlocker(function () {
                    var begin_button = document.getElementById("begin-button");
                    if (begin_button) {
                        begin_button.setAttribute('disabled', 'disabled');
                        window.alert(
                            'You appear to have an ad blocker enabled. ' +
                            'This experiment is incompatible ' +
                            'with most ad blockers. Please pause your ad ' +
                            'blocker or disable it for this domain and ' +
                            'reload the page.'
                        );
                    }
                });
            </script>
        {% endblock %}
    </body>
</html>
