{% extends "layout.html" %}

{% block body %}
    <div class="main_div">
        <div>
            <h1>Thank you!</h1>
            <hr>
            <div>
                <h1>Your feedback has been submitted.</h1>
                {% if completed %}
                <p>The HIT has been marked as completed.
                    You should be compensated automatically.</p>
                <p>To enquire about compensation, please contact the researcher at
                    <a href="mailto:{{contact_address}}">{{contact_address}}</a>
                        and include your HIT ID: {{hit_id}}</p>
                {% else %}
                <p><b>We were not able to mark this HIT as completed.</b></p>
                <p>Please contact the researcher at
                    <a href="mailto:{{contact_address}}">{{contact_address}}</a>
                        and include your HIT ID: {{hit_id}}</p>
                {% endif %}
            </div>
        </div>
    </div>
{% endblock %}
