{% extends "base/dashboard.html" %}

{% block body %}
<h1>Experiment Configuration</h1>
<p>The active <a href="https://dallinger.readthedocs.io/en/latest/configuration.html">configuration</a>,
    assembled from your configuration files, Dallinger defaults, and any explicit overrides.
</p>
<table class="table table-sm table-striped">
    <thead class="thead-dark">
        <tr><th>Key</th><th>Active Value</th></tr>
    </thead>
    <tbody>
    {% for key, value in configuration %}
        <tr><th>{{ key }}</th><td>{{ value }}</td></tr>
    {% endfor %}
    </tbody>
</table>
{% endblock %}