{% extends "layout.html" %}

<!--
  The consent.html displays the text of your IRB-approved
  consent form.  Even if you are not required to provide
  a consent form, it is helpful to use this form to
  describe what people can do if an error comes up, etc...

-->

{% block body %}
    <div class="main_div">
        <noscript>
            <h1>Warning: Javascript is not working on your browser</h1>
            <p>The study will not work unless Javascript is enabled on your browser.</p>
            <p>Please enable Javascript then refresh this page.<p>
        </noscript>
        <h1>We need your consent to proceed</h1>
        <hr>
        <div class="consent_div text-justify"><small>
            {% block consent_text %}
            <p>This research is being conducted by XXXX, a postdoctoral research fellow, and XXXX, a faculty member, at XXXX. This study takes 10 minutes.</p>

            <p>If you agree to take part in the research, you will complete a study that involves reading some text, remembering it as well as you can, and then answering some questions about it. All of the information we obtain during the research will be kept confidential, and not associated with your name in any way, however, while the study is running it will be associated with your MTurk worker id. Once the study is complete we will replace your worker id with a random string.</p>

            <p>Your participation in this research is voluntary. You are free to refuse to take part, and you may stop taking part at any time. You are free to discontinue participation in this study at any time with no penalty.</p>

            <p>If you have any questions about this research, do not hesitate to contact XXXX at XXXX. If you have any questions about your rights or treatment as a participant in this research project, please contact XXXX's XXXX at XXXXX or XXXX@XXXX.edu.</p>

            <p>By consenting to participate, you acknowledge that you are 18 years or older, have read this consent form, agree to its contents, and agree to take part in this research. If you do not wish to consent, close this page and return the HIT on Mechanical Turk.</p>
            {% endblock consent_text %}
            <hr>
            <h4>Do you understand and consent to these terms?</h4>
            <br>
            {% block consent_button %}
            <button type="button" class="btn btn-primary btn-lg" id="consent" onClick="window.location='/instructions/instruct-ready?{{query_string | safe}}';" style="float: left;">I agree
            </button>
            {% endblock consent_button %}
            {% block reject_button %}
            <button type="button" class="btn btn-danger btn-lg" onClick="self.close();" style="float: right;">No thanks (exit HIT)
            </button>
            {% endblock reject_button %}
        </div>
    </div>
{% endblock %}
